#include "Main.h"

#include "../Framework2/Util/TexturePacker.h"

void MainApp::initMenu(){
	Menu *menu = menuSystem->getMainMenu();
	menu->addMenuItem("Draw static lights: ", &drawStaticLights, INPUT_BOOL);
	menu->addMenuItem("Draw dynamic light: ", &drawDynamicLight, INPUT_BOOL);
	menu->addMenuItem("Use clip planes: ", &useClipPlanes, INPUT_BOOL);

	App::initMenu();
}

void MainApp::resetCamera(){
	position = vec3(976, -140, -976);
	wx = 0.15f;
	wy = 0.68f;
	wz = 0;
}

void MainApp::processMovement(const vec3 &newPosition){
	vec3 pos = newPosition;

	coll.pushSphere(pos, 48);

	position = pos;
}

bool MainApp::init(){
	unsigned int i, j;

	drawDynamicLight = true;
	drawStaticLights = true;
	useClipPlanes = true;

	model = new Direct3DModel();
	model->loadFromFile("../Models/PillarRoom/Map2.hmdl");


	staticLights[0].pos = vec3(-1450, 690, -300);
	staticLights[0].size = 100;
	staticLights[0].radius = 2300;
	staticLights[0].color = vec3(1, 0.6f, 0);

	staticLights[1].pos = vec3(230, -750, 1600);
	staticLights[1].size = 50;
	staticLights[1].radius = 1500;
	staticLights[1].color = vec3(1, 0.4f, 0.1f);

	staticLights[2].pos = vec3(550, 1150, 0);
	staticLights[2].size = 50;
	staticLights[2].radius = 1500;
	staticLights[2].color = vec3(1, 0.3f, 0);


	for (i = 0; i < model->getBatchCount(); i++){
		model->getBatch(i)->insertAttribute(ATT_TEXCOORD, ATT_FLOAT, 2, 4);
	}

	TexturePacker texPacker;
	for (i = 0; i < model->getBatchCount(); i++){
		Batch *batch = model->getBatch(i);
		Vertex *src = (Vertex *) batch->getVertices();
		unsigned short *inds = (unsigned short *) batch->getIndices();

		for (j = 0; j < batch->getIndexCount(); j += 6){
			float w = length(src[inds[j + 1]].pos - src[inds[j]].pos) / 28;
			float h = length(src[inds[j + 5]].pos - src[inds[j]].pos) / 28;

			if (w < 8) w = 8;
			if (h < 8) h = 8;

			unsigned int tw = (int) w;
			unsigned int th = (int) h;

			tw = (tw + 4) & ~7;
			th = (th + 4) & ~7;

			texPacker.addRectangle(tw, th);
		}
	}

	unsigned int tw = 512;
	unsigned int th = 256;

	if (!texPacker.assignCoords(&tw, &th)){
		MessageBox(NULL, "Lightmap too small", "Error", MB_OK);
		return false;
	}


	tw = getUpperPowerOfTwo(tw);
	th = getUpperPowerOfTwo(th);


	Set <PolyData *> polyData;

	unsigned int index = 0;
	for (i = 0; i < model->getBatchCount(); i++){
		Batch *batch = model->getBatch(i);
		Vertex *src = (Vertex *) batch->getVertices();
		unsigned short *inds = (unsigned short *) batch->getIndices();

		for (j = 0; j < batch->getIndexCount(); j += 6){
			TextureRectangle *rect = texPacker.getRectangle(index);

			float x0 = float(rect->x + 0.5f) / tw;
			float x1 = float(rect->x + rect->width - 0.5f) / tw;
			float y0 = float(rect->y + 0.5f) / th;
			float y1 = float(rect->y + rect->height - 0.5f) / th;

			src[inds[j    ]].texCoord2 = vec2(x0, y0);
			src[inds[j + 1]].texCoord2 = vec2(x1, y0);
			src[inds[j + 2]].texCoord2 = vec2(x1, y1);
			src[inds[j + 5]].texCoord2 = vec2(x0, y1);

			PolyData *pd = new PolyData;
			pd->v0 = src[inds[j]].pos;
			pd->vDir0 = src[inds[j + 1]].pos - pd->v0;
			pd->vDir1 = src[inds[j + 5]].pos - pd->v0;
			pd->t0 = vec2(x0, y0);
			pd->tDir0 = vec2(x1 - x0, 0);
			pd->tDir1 = vec2(0, y1 - y0);
			polyData.add(pd);

			coll.addTriangle(src[inds[j]].pos, src[inds[j + 1]].pos, src[inds[j + 2]].pos, pd);
			coll.addTriangle(src[inds[j + 3]].pos, src[inds[j + 4]].pos, src[inds[j + 5]].pos, pd);

			index++;
		}
	}
	coll.build(3, 1, 50);



	/*****	Lightmap generation code *****/
/*
	vec3 samples[SAMPLE_COUNT];
	for (unsigned int x = 0; x < SAMPLE_COUNT; x++){
		vec3 v;
		do {
			v = vec3(float(rand()), float(rand()), float(rand())) * (2.0f / RAND_MAX) - 1.0f;
		} while (dot(v, v) > 1);
		samples[x] = v;
	}


	for (unsigned int lm = 0; lm < STATIC_LIGHT_COUNT; lm++){
		float invRadius = 1.0f / staticLights[lm].radius;

		unsigned char *lMap = new unsigned char[tw * th];
		memset(lMap, 0, tw * th);

		index = 0;
		for (i = 0; i < model->getBatchCount(); i++){
			Batch *batch = model->getBatch(i);
			Vertex *src = (Vertex *) batch->getVertices();
			unsigned short *inds = (unsigned short *) batch->getIndices();

			for (j = 0; j < batch->getIndexCount(); j += 6){
				TextureRectangle *rect = texPacker.getRectangle(index);

				vec3 dirS = (src[inds[j + 1]].pos - src[inds[j]].pos);
				vec3 dirT = (src[inds[j + 5]].pos - src[inds[j]].pos);

				vec3 pos = src[inds[j]].pos;
				//vec3 normal = src[inds[j]].normal;

				for (int t = 0; t < int(rect->height); t++){
					for (int s = 0; s < int(rect->width); s++){
						vec3 samplePos = pos + saturate((float(s) / (rect->width - 1))) * dirS + saturate((float(t) / (rect->height - 1))) * dirT;

						coll.pushSphere(samplePos, 5);

						float light = 0.0f;
						for (unsigned int k = 0; k < SAMPLE_COUNT; k++){
							if (!coll.intersect(samplePos, staticLights[lm].pos + staticLights[lm].size * samples[k])) light += 1.0f / SAMPLE_COUNT;
						}
	
						vec3 lightVec = invRadius * (staticLights[lm].pos - samplePos);
						light *= saturate(1 - dot(lightVec, lightVec));
						//light /= (1.0f + 0.0000004f * dot(lightVec, lightVec));

						lMap[((rect->y + t) * tw + (rect->x + s))] = (unsigned char) (255.0f * light);
					}
				}

				index++;
			}
		}

		Image image;
		image.loadFromMemory(lMap, tw, th, FORMAT_I8, true, false);
		String fileName;
		fileName.sprintf("LightMap%d.dds", lm);
		image.saveImage(fileName);
	}
*/

	for (i = 0; i < model->getBatchCount(); i++){
		model->getBatch(i)->fixTJunctions();
	}

	return true;
}

bool MainApp::exit(){
	delete model;
	return true;
}

bool MainApp::load(){
	if (caps.PixelShaderVersion < D3DPS_VERSION(2,0) || caps.VertexShaderVersion < D3DVS_VERSION(1,1)){
		String error("This demo requires pixel shader 2.0 and vertex shader 1.1.\n"
					 "Your card/driver only supports pixel shader ");
		error.sprintf("%d.%d", (caps.PixelShaderVersion  >> 8) & 255, caps.PixelShaderVersion  & 255);
		error += " and vertex shader ";
		error.sprintf("%d.%d", (caps.VertexShaderVersion >> 8) & 255, caps.VertexShaderVersion & 255);

		addToLog(error);
		return false;
	}

	setDefaultFont("../Textures/Fonts/Future.font", "../Textures/Fonts/Future.dds");

	if ((distance = renderer->addRenderCubemap(RT_SIZE, FORMAT_I32F)) == TEXTURE_NONE) return false;
	if ((distanceDepth = renderer->addRenderDepth(RT_SIZE, RT_SIZE, FORMAT_D16)) == TEXTURE_NONE) return false;

	if ((shadow = renderer->addRenderTexture(RT_SIZE, RT_SIZE, FORMAT_RGBA8, false, true)) == TEXTURE_NONE) return false;
	if ((shadowDepth = renderer->addRenderDepth(RT_SIZE, RT_SIZE, FORMAT_D16, true)) == TEXTURE_NONE) return false;

	if ((distanceShader = renderer->addShader("distance.shd", DEFINE_STR(INV_LIGHT_RANGE))) == SHADER_NONE) return false;
	if ((shadowShader   = renderer->addShader("shadow.shd",   DEFINE_STR(INV_LIGHT_RANGE))) == SHADER_NONE) return false;
	if ((lightingShader = renderer->addShader("lighting.shd", DEFINE_STR(INV_LIGHT_RANGE))) == SHADER_NONE) return false;
	if ((staticLightingShader = renderer->addShader("staticLighting.shd")) == SHADER_NONE) return false;
	if ((ambientLightingShader = renderer->addShader("ambientLighting.shd")) == SHADER_NONE) return false;

	if ((base[0] = renderer->addTexture("../Textures/Tx_imp_wall_01_small.dds")) == TEXTURE_NONE) return false;
	if ((bump[0] = renderer->addTexture("../Textures/Tx_imp_wall_01Bump.png", TEX_NORMALMAP)) == TEXTURE_NONE) return false;

	if ((base[1] = renderer->addTexture("../Textures/wood056.dds")) == TEXTURE_NONE) return false;
	if ((bump[1] = renderer->addTexture("../Textures/wood056Bump.png", TEX_NORMALMAP)) == TEXTURE_NONE) return false;

	if ((base[2] = renderer->addTexture("../Textures/stone08.dds")) == TEXTURE_NONE) return false;
	if ((bump[2] = renderer->addTexture("../Textures/stone08Bump.png", TEX_NORMALMAP)) == TEXTURE_NONE) return false;

	if ((lightSpot = renderer->addTexture("../Textures/Particle.png")) == TEXTURE_NONE) return false;

	for (unsigned int i = 0; i < STATIC_LIGHT_COUNT; i++){
		String fileName;
		fileName.sprintf("LightMap%d.dds", i);
		
		if ((lightMaps[i] = renderer->addTexture(fileName)) == TEXTURE_NONE) return false;
	}

	model->uploadToVertexBuffer(dev);

	return true;
}

bool MainApp::unload(){
	model->freeVertexBuffer();
	return true;
}

void MainApp::drawLight(const Light &light){
	renderer->setTextures(lightSpot);
	renderer->setBlending(ONE, ONE);
	renderer->apply();

	vec3 dx(modelView.elem[0][0], modelView.elem[0][1], modelView.elem[0][2]);
	vec3 dy(modelView.elem[1][0], modelView.elem[1][1], modelView.elem[1][2]);

	struct Vertex {
		vec3 position;
		unsigned int color;
		vec2 texCoord;
	};

	Vertex vertices[4];

	vertices[0].position = light.pos + light.size * (-dx + dy);
	vertices[1].position = light.pos + light.size * ( dx + dy);
	vertices[2].position = light.pos + light.size * ( dx - dy);
	vertices[3].position = light.pos + light.size * (-dx - dy);

	vertices[0].texCoord = vec2(0, 0);
	vertices[1].texCoord = vec2(1, 0);
	vertices[2].texCoord = vec2(1, 1);
	vertices[3].texCoord = vec2(0, 1);

	vertices[0].color = vertices[1].color = vertices[2].color = vertices[3].color = toBGRA(vec4(light.color, 0.0f));

	dev->SetFVF(D3DFVF_XYZ | D3DFVF_DIFFUSE | D3DFVF_TEXCOORDSIZE2(0) | (1 << D3DFVF_TEXCOUNT_SHIFT));
	dev->DrawPrimitiveUP(D3DPT_TRIANGLEFAN, 2, vertices, sizeof(Vertex));
}

void MainApp::drawDistance(const Light &light){
	renderer->setShader(distanceShader);
	renderer->apply();

	renderer->changeShaderConstant3f("lightPos", light.pos);

	mat4 proj = projectionMatrixX(PI / 2, 1, 1, light.radius);

	for (int face = 0; face < 6; face++){
		renderer->changeRenderTarget(distance, distanceDepth, face);
		dev->Clear(0, NULL, D3DCLEAR_ZBUFFER, D3DCOLOR_RGBA(0,0,0,0), 1.0f, 0);

		mat4 view = cubemapMatrix(face);
		view.translate(-light.pos);
		renderer->changeShaderConstant4x4f("mvp", proj * view);

		model->draw(dev);
	}
}

void MainApp::drawShadow(const Light &light){
	renderer->changeRenderTarget(shadow, shadowDepth);
	dev->Clear(0, NULL, D3DCLEAR_ZBUFFER, D3DCOLOR_RGBA(0,0,0,0), 1.0f, 0);

	renderer->setShader(shadowShader);
	renderer->setTexture("Dist", distance);
	renderer->apply();

	renderer->changeShaderConstant4x4f("mvp", projection * modelView);
	renderer->changeShaderConstant3f("lightPos", light.pos);
	renderer->changeShaderConstant3f("camPos", position);

	model->draw(dev);
	
	renderer->changeToDefaultRenderTarget();
}

void MainApp::setupClipPlanes(const mat4 &itmvp, const vec3 &pos, const float radius){
	vec4 plane0( 1,  0,  0, radius - pos.x);
	vec4 plane1(-1,  0,  0, radius + pos.x);
	vec4 plane2( 0,  1,  0, radius - pos.y);
	vec4 plane3( 0, -1,  0, radius + pos.y);
	vec4 plane4( 0,  0,  1, radius - pos.z);
	vec4 plane5( 0,  0, -1, radius + pos.z);

	dev->SetClipPlane(0, itmvp * plane0);
	dev->SetClipPlane(1, itmvp * plane1);
	dev->SetClipPlane(2, itmvp * plane2);
	dev->SetClipPlane(3, itmvp * plane3);
	dev->SetClipPlane(4, itmvp * plane4);
	dev->SetClipPlane(5, itmvp * plane5);
}

bool MainApp::drawFrame(){
    dev->Clear(0, NULL, D3DCLEAR_TARGET | D3DCLEAR_ZBUFFER, D3DCOLOR_RGBA(64, 64, 64, 0), 1.0f, 0);

	projection = projectionMatrixX(1.5f, float(height) / float(width), 1, 6000);
	dev->SetTransform(D3DTS_PROJECTION, (D3DMATRIX *) (const float *) transpose(projection));

	modelView = rotateZXY(-wx, -wy, -wz);
	modelView.translate(-position);
	dev->SetTransform(D3DTS_VIEW, (D3DMATRIX *) (const float *) transpose(modelView));

	mat4 mvp = projection * modelView;

	dev->SetRenderState(D3DRS_CULLMODE, D3DCULL_CCW);

	for (unsigned int i = 0; i < model->getBatchCount(); i++){
		renderer->setShader(ambientLightingShader);
		renderer->setTexture("Base", base[i]);
		renderer->apply();

		renderer->changeShaderConstant4x4f("mvp", mvp);

		((Direct3DBatch *) model->getBatch(i))->draw(dev);
	}


	float bias  = -0.00001f;
	float sBias = -0.00001f;
	mat4 itmvp = !transpose(mvp);

	if (drawStaticLights){
		if (useClipPlanes){
			dev->SetRenderState(D3DRS_CLIPPLANEENABLE, D3DCLIPPLANE0 | D3DCLIPPLANE1 | D3DCLIPPLANE2 | D3DCLIPPLANE3 | D3DCLIPPLANE4 | D3DCLIPPLANE5);
			dev->SetRenderState(D3DRS_DEPTHBIAS, *(DWORD *) &bias);
			dev->SetRenderState(D3DRS_SLOPESCALEDEPTHBIAS, *(DWORD *) &sBias);
		}
		
		for (unsigned int j = 0; j < STATIC_LIGHT_COUNT; j++){
			if (useClipPlanes) setupClipPlanes(itmvp, staticLights[j].pos, staticLights[j].radius);

			for (unsigned int i = 0; i < model->getBatchCount(); i++){
				renderer->setShader(staticLightingShader);
				renderer->setTexture("Base", base[i]);
				renderer->setTexture("Bump", bump[i]);
				renderer->setTexture("LightMap", lightMaps[j]);
				renderer->setBlending(ONE, ONE);
				renderer->setMask(COLOR);
				renderer->apply();

				renderer->changeShaderConstant4x4f("mvp", mvp);
				renderer->changeShaderConstant3f("lightPos", staticLights[j].pos);
				renderer->changeShaderConstant4f("lightColor", vec4(staticLights[j].color, 0));

				((Direct3DBatch *) model->getBatch(i))->draw(dev);
			}
		}

		if (useClipPlanes){
			dev->SetRenderState(D3DRS_CLIPPLANEENABLE, 0);
			dev->SetRenderState(D3DRS_DEPTHBIAS, 0);
			dev->SetRenderState(D3DRS_SLOPESCALEDEPTHBIAS, 0);
		}
	}


	if (drawDynamicLight){
		dynamicLight.pos = vec3(800 * cosf(time), 380 * sinf(time / 3) - 230, 800 * sinf(time));
		dynamicLight.size = 120;
		dynamicLight.radius = LIGHT_RANGE;
		dynamicLight.color = vec3(1, 1, 1);

		static vec4 offsets[] = {
			vec4( 0.8f, -1.0f, 0.0f, 0.0f), //   . . . . . | . . . X .
			vec4(-0.2f, -0.8f, 0.0f, 0.0f), //   . . . . X | . . . . .
			vec4( 0.2f, -0.6f, 0.0f, 0.0f), //   . . . . . | X . . . .
			vec4( 1.0f, -0.4f, 0.0f, 0.0f), //   . . . . . | . . . . X
			vec4(-0.6f, -0.2f, 0.0f, 0.0f), //   . . X . . | . . . . .
			                                //   ----------X----------
			vec4( 0.6f,  0.2f, 0.0f, 0.0f), //   . . . . . | . . X . .
			vec4(-1.0f,  0.4f, 0.0f, 0.0f), //   X . . . . | . . . . .
			vec4( 0.4f,  0.6f, 0.0f, 0.0f), //   . . . . . | . X . . .
			vec4(-0.4f,  0.8f, 0.0f, 0.0f), //   . . . X . | . . . . .
			vec4(-0.8f,  1.0f, 0.0f, 0.0f), //   . X . . . | . . . . .
		};
		
		drawDistance(dynamicLight);

		if (useClipPlanes){
			dev->SetRenderState(D3DRS_CLIPPLANEENABLE, D3DCLIPPLANE0 | D3DCLIPPLANE1 | D3DCLIPPLANE2 | D3DCLIPPLANE3 | D3DCLIPPLANE4 | D3DCLIPPLANE5);
			dev->SetRenderState(D3DRS_DEPTHBIAS, *(DWORD *) &bias);
			dev->SetRenderState(D3DRS_SLOPESCALEDEPTHBIAS, *(DWORD *) &sBias);

			setupClipPlanes(itmvp, dynamicLight.pos, dynamicLight.radius);
		}

		drawShadow(dynamicLight);

		
		for (unsigned int i = 0; i < model->getBatchCount(); i++){
			renderer->setShader(lightingShader);
			renderer->setTexture("Base", base[i]);
			renderer->setTexture("Bump", bump[i]);
			renderer->setTexture("Shadow", shadow);
			renderer->setBlending(ONE, ONE);
			renderer->setMask(COLOR);
			renderer->apply();

			renderer->changeShaderConstant4x4f("mvp", projection * modelView);
			renderer->changeShaderConstant3f("lightPos", dynamicLight.pos);
			renderer->changeShaderConstant3f("camPos", position);
			renderer->changeShaderConstant4f("lightColor", vec4(dynamicLight.color, 0));

			renderer->changeShaderConstant1f("invWidth",  1.0f / width);
			renderer->changeShaderConstant1f("invHeight", 1.0f / height);

			renderer->changeShaderConstantArray4f("offsets", offsets, sizeof(offsets) / sizeof(offsets[0]));

			((Direct3DBatch *) model->getBatch(i))->draw(dev);
		}

		if (useClipPlanes){
			dev->SetRenderState(D3DRS_CLIPPLANEENABLE, 0);
			dev->SetRenderState(D3DRS_DEPTHBIAS, 0);
			dev->SetRenderState(D3DRS_SLOPESCALEDEPTHBIAS, 0);
		}
	}

	if (drawStaticLights){
		for (unsigned int j = 0; j < STATIC_LIGHT_COUNT; j++){
			drawLight(staticLights[j]);
		}
	}
	if (drawDynamicLight) drawLight(dynamicLight);

	return true;
}

App *app = new MainApp();
